// 2-port data_mux
//
// Prioritises Port A for bi-directional data transfer,
// holds requests from Port x for x cycles whilst another
// request is in progress (x cycles = GPU RAM delay).
//
// All requests should take no longer than the specified
// number of cycles to read data from the GPU RAM, but in
// any case be quicker than the reading device attached
// to this mux unit - i.e. the Z80 or RS232 interfaces.

module data_mux_v2 (

   // general inputs
   input wire        clk,
   input wire        reset,
   
   // gpu inputs
   input wire [7:0]  gpu_data_in,   // data is valid here once gpu_data_in pulses HIGH
   
   // inputs Port A - Z80
   input wire        wr_ena_a,
   input wire        rd_req_a,
   input wire [19:0] address_a,
   input wire [7:0]  data_in_a,
   
   
   // inputs Port B - RS232
   input wire        wr_ena_b,
   input wire        rd_req_b,
   input wire [19:0] address_b,
   input wire [7:0]  data_in_b,
   
   // gpu outputs
   output reg        gpu_wr_ena,    // output pulses high for 1 clock when a write request takes place
   output reg [19:0] gpu_address,
   output reg [7:0]  gpu_data_out,
   
   // outputs Port A
   output reg        gpu_rd_rdy_a,  // one-CLK HIGH pulse here indicates valid data on data_out_a
   output wire [7:0] data_out_a,
   
   // outputs Port B
   output reg        gpu_rd_rdy_b,  // one-CLK HIGH pulse here indicates valid data on data_out_b
   output wire [7:0] data_out_b

);

parameter  READ_CLOCK_CYCLES = 2 ;

reg [1:0] mux_priority ;
reg [9:0] rd_req_dlya, rd_req_dlyb ;

assign data_out_a   =  gpu_data_in ;  // with this line, it is the responsibility of the next module to latch the data when the gpu_rd_rdy_a is high
assign data_out_b   =  gpu_data_in ;  // with this line, it is the responsibility of the next module to latch the data when the gpu_rd_rdy_b is high


wire        F_wr_ena_a;
wire        F_rd_req_a;
wire [19:0] F_address_a;
wire [7:0]  F_data_in_a;
wire        cmd_a_next, cmd_a_rdy;

wire        F_wr_ena_b;
wire        F_rd_req_b;
wire [19:0] F_address_b;
wire [7:0]  F_data_in_b;
wire        cmd_b_next, cmd_b_rdy;

FIFO_3word_0_latency input_cmd_fifo_1 (     // Zero Latency Command buffer.
                      .clk(clk),                                                    // CLK input
                      .reset(reset),                                                // reset FIFO

                      .shift_in        ( rd_req_a || wr_ena_a ),                    // load a word into the FIFO.
                      .shift_out       ( cmd_a_next ),                              // shift data out of the FIFO.
                      .data_in         ( {rd_req_a,wr_ena_a,address_a,data_in_a} ), // data word input.

                      .fifo_not_empty  ( cmd_a_rdy ),                                      // High when there is data available.
                      .fifo_full       (),                                                 // High when the FIFO is full.
                      .data_out        ( {F_rd_req_a,F_wr_ena_a,F_address_a,F_data_in_a} ) // FIFO data word output
                       );
	defparam
		input_cmd_fifo_1.bits         = (1+1+20+8),  // The number of bits containing the command.
		input_cmd_fifo_1.zero_latency = 1;

FIFO_3word_0_latency input_cmd_fifo_2 (     // Zero Latency Command buffer.
                      .clk(clk),                                                    // CLK input
                      .reset(reset),                                                // reset FIFO

                      .shift_in( rd_req_b || wr_ena_b ),                            // load a word into the FIFO.
                      .shift_out( cmd_b_next ),                                     // shift data out of the FIFO.
                      .data_in         ( {rd_req_b,wr_ena_b,address_b,data_in_b} ), // data word input.

                      .fifo_not_empty  ( cmd_b_rdy ),                                      // High when there is data available.
                      .fifo_full       (),                                                 // High when the FIFO is full.
                      .data_out        ( {F_rd_req_b,F_wr_ena_b,F_address_b,F_data_in_b} ) // FIFO data word output
                       );
	defparam
		input_cmd_fifo_2.bits         = (1+1+20+8),  // The number of bits containing the command.
		input_cmd_fifo_2.zero_latency = 1;


assign cmd_a_next = (cmd_a_rdy && (~cmd_b_rdy || mux_priority!=2'd0) );
assign cmd_b_next = (cmd_b_rdy && (~cmd_a_rdy || mux_priority!=2'd1) );

always @ (posedge clk) begin

	gpu_rd_rdy_a   <= rd_req_dlya[READ_CLOCK_CYCLES-1]; // needs to be high when 2 clock cycles has passed since address_a was sent out
	gpu_rd_rdy_b   <= rd_req_dlyb[READ_CLOCK_CYCLES-1]; // needs to be high when 2 clock cycles has passed since address_b was sent out
	rd_req_dlya[9:1] <= rd_req_dlya[8:0] ; // delay the read request by the correct amount of clocks matching the 
	rd_req_dlyb[9:1] <= rd_req_dlyb[8:0] ; // delay the read request by the correct amount of clocks matching the 


if ( cmd_a_next ) begin
		gpu_wr_ena          <=  F_wr_ena_a && cmd_a_rdy  ;
		gpu_address         <=  F_address_a ;
		gpu_data_out        <=  F_data_in_a ;
		rd_req_dlya[0]      <=  F_rd_req_a && cmd_a_rdy ;  // internally hold the read request
		mux_priority        <=  2'd0;
	end else rd_req_dlya[0]	<=  1'b0;

if ( cmd_b_next ) begin
		gpu_wr_ena          <=  F_wr_ena_b && cmd_b_rdy  ;
		gpu_address         <=  F_address_b ;
		gpu_data_out        <=  F_data_in_b ;
		rd_req_dlyb[0]      <=  F_rd_req_b && cmd_b_rdy ;  // internally hold the read request
		mux_priority        <=  2'd1;
	end else rd_req_dlyb[0]	<=  1'b0;


end // always

endmodule
